"""Tests a trained re-implementation of the FAME3 model.

This script saves the test metrics to a text file.
This script saves the per-atom predictions to a CSV file.
This script performs bootstrapping to estimate the uncertainty in the metrics.\
    The number of bootstraps can be set by changing the NUM_BOOTSTRAPS variable. Default is 1000.
The radius of the atom environment is not part of the hyperparameter search, \
    but can be set by changing the radius argument. Default is 5.
The decision threshold can be changed by changing the threshold argument. Default is 0.3.
The script also computes FAME scores if the -fs flag is set.
"""

import argparse
import csv
import glob
import os
import sys
from datetime import datetime
from statistics import mean, stdev

import numpy as np
import pandas as pd
from joblib import load

from fame3r import FAMEDescriptors, FAMEScores, compute_metrics

NUM_BOOTSTRAPS = 1000


def parse_arguments() -> argparse.Namespace:
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description="Tests a trained re-implementation of the FAME3 model."
    )

    parser.add_argument(
        "-i",
        dest="input_file",
        required=True,
        metavar="<Input data file>",
        help="Input data file",
    )
    parser.add_argument(
        "-m",
        dest="model_folder",
        required=True,
        metavar="<Model folder>",
        help="Model folder",
    )
    parser.add_argument(
        "-o",
        dest="out_folder",
        required=True,
        metavar="<Output folder>",
        help="Output location",
    )
    parser.add_argument(
        "-r",
        dest="radius",
        required=False,
        metavar="<radius>",
        default=5,
        help="Max. atom environment radius in number of bonds",
        type=int,
    )
    parser.add_argument(
        "-t",
        dest="threshold",
        required=False,
        metavar="<binary decision threshold>",
        default=0.3,
        help="Binary decision threshold",
        type=float,
    )
    parser.add_argument(
        "-fs",
        dest="compute_fame_scores",
        action="store_true",
        help="Compute FAME scores (optional)",
    )

    parse_args = parser.parse_args()

    return parse_args


def main():
    """Application entry point."""
    start_time = datetime.now()

    args = parse_arguments()
    print(f"Radius: {args.radius}")

    if not os.path.exists(args.out_folder):
        os.makedirs(args.out_folder)
        print("The new output folder is created.")

    print("Computing descriptors...")
    descriptors_generator = FAMEDescriptors(args.radius)
    mol_num_ids, mol_ids, atom_ids, som_labels, descriptors = descriptors_generator.compute_fame_descriptors(
        args.input_file, args.out_folder, has_soms=True
    )

    print(f"Testing data: {len(set(mol_num_ids))} molecules")

    print("Loading model...")
    model_file = os.path.join(args.model_folder, "model.joblib")
    clf = load(model_file)

    print("Testing model...")
    y_prob = clf.predict_proba(descriptors)[:, 1]
    y_pred = (y_prob > args.threshold).astype(int)

    fame_scores = None
    if args.compute_fame_scores:
        print("Computing FAME scores...")
        csv_files = glob.glob(os.path.join(args.model_folder, "*descriptors.csv"))
        if len(csv_files) == 1:
            train_descriptors_df = pd.read_csv(csv_files[0])
        else:
            raise FileNotFoundError(
                f"Expected one CSV file ending with 'descriptors.csv', but found {len(csv_files)}."
            )
        train_descriptors = train_descriptors_df.drop(
            columns=["mol_num_id", "mol_id", "atom_id", "som_label"]
        ).values
        fame_scores_generator = FAMEScores(train_descriptors)
        fame_scores = fame_scores_generator.compute_fame_scores(descriptors)

    print("Computing metrics...")
    metrics: dict[str, list[float]] = {
        "AUROC": [],
        "Average precision": [],
        "F1": [],
        "MCC": [],
        "Precision": [],
        "Recall": [],
        "Top-2 correctness rate": [],
    }

    for i in range(NUM_BOOTSTRAPS):
        print(f"Bootstrap {i + 1}...")
        sampled_mol_num_ids = np.random.choice(
            list(set(mol_num_ids)), len(set(mol_num_ids)), replace=True
        )
        mask = np.zeros(len(mol_num_ids), dtype=bool)
        for mol_num_id in sampled_mol_num_ids:
            mask = mask | (mol_num_ids == mol_num_id)
        print("Computing metrics...")
        assert som_labels is not None, "SOM labels are missing"
        (
            auroc,
            average_precision,
            f1,
            mcc,
            precision,
            recall,
            top2,
        ) = compute_metrics(
            som_labels[mask], y_prob[mask], y_pred[mask], mol_num_ids[mask]
        )

        metrics["AUROC"].append(auroc)
        metrics["Average precision"].append(average_precision)
        metrics["F1"].append(f1)
        metrics["MCC"].append(mcc)
        metrics["Precision"].append(precision)
        metrics["Recall"].append(recall)
        metrics["Top-2 correctness rate"].append(top2)

    metrics_file = os.path.join(args.out_folder, "test_metrics.txt")
    with open(metrics_file, "w", encoding="UTF-8") as file:
        for metric, scores in metrics.items():
            file.write(
                f"{metric}: {round(mean(scores), 4)} +/- {round(stdev(scores), 4)}\n"
            )
    print(f"Metrics saved to {metrics_file}")

    assert som_labels is not None, "SOM labels are missing"
    predictions_file = os.path.join(args.out_folder, "predictions.csv")
    with open(predictions_file, "w", encoding="UTF-8", newline="") as file:
        writer = csv.writer(file)
        headers = ["mol_id", "atom_id", "y_prob", "y_pred", "y_true"]
        if args.compute_fame_scores:
            headers.append("fame_score")
        writer.writerow(headers)

        for i in range(len(mol_ids)):
            row = [
                mol_ids[i],
                atom_ids[i],
                y_prob[i],
                y_pred[i],
                som_labels[i],
            ]
            if args.compute_fame_scores:
                row.append(fame_scores[i])
            writer.writerow(row)

    print(f"Predictions saved to {predictions_file}")
    print("Finished in:", datetime.now() - start_time)

    sys.exit(0)


if __name__ == "__main__":
    main()


"""Module for computing Fame Scores."""

from statistics import mean

import numpy as np
from CDPL import Descr, Math


class FAMEScores:
    """Class for computing Fame Scores."""

    def __init__(self, train_descriptors, num_nearest_neighbors=3):
        """
        Initialize the FameScores class.

        Args:
            data (np.ndarray): 2D array of descriptors (rows: samples, columns: descriptors).
            num_nearest_neighbors (int): Number of nearest neighbors to consider.
        """
        self.train_descriptors = train_descriptors
        self.num_nearest_neighbors = num_nearest_neighbors

        # Initialize bulk similarity calculator for
        # calculating the similarity between
        # double precision floating point descriptor vectors.
        # Default similarity measure is Tanimoto.
        self.sim_calc = Descr.DVectorBulkSimilarityCalculator()

        # Add descriptors to the similarity calculator
        for i in range(self.train_descriptors.shape[0]):
            descriptor = Math.DVector(self.train_descriptors[i, :])

            # Check if the descriptor is already in the calculator
            # to avoid duplicates.
            if descriptor in self.sim_calc:
                continue

            self.sim_calc.addDescriptor(descriptor)

    def compute_fame_scores(self, data):
        """
        Compute the fame scores based on the provided data.

        Args:
            data (np.ndarray): 2D array of descriptors (rows: samples, columns: descriptors).

        Returns:
            np.ndarray: 1D array of fame scores.
        """
        fame_scores = []

        # Iterate over each descriptor in the data
        for i in range(data.shape[0]):
            query = Math.DVector(data[i, :])

            # Calculate similarities and order results
            # (enabled by sec. arg. == True) by descending value
            self.sim_calc.calculate(query, True)
            temp_fame_scores = []
            for k in range(self.num_nearest_neighbors):
                res = self.sim_calc[k]
                temp_fame_scores.append(res[1])
                # res[0] is the index of the datapoint and
                # res[1] is the similarity score
            fame_scores.append(round(mean(temp_fame_scores),2))

        # Return the fame scores as a numpy array
        return np.array(fame_scores, dtype=float)



"""Module for computing FAME descriptors for a given molecule."""

import ast
import csv
import os
import sys
from pathlib import Path

import numpy as np
from CDPL import Chem, ForceField, MolProp

SYBYL_ATOM_TYPE_IDX_CDPKIT = [
    1,
    2,
    3,
    4,
    6,
    7,
    8,
    9,
    10,
    11,
    12,
    13,
    14,
    15,
    18,
    19,
    20,
    21,
    22,
    23,
    24,
    38,
    47,
    48,
    49,
    54,
]


class MoleculeProcessor:
    """Class for processing molecules and extracting structure data."""

    @staticmethod
    def perceive_mol(mol: Chem.Molecule) -> None:
        """Perceive various molecular properties and set corresponding attributes."""
        Chem.calcImplicitHydrogenCounts(mol, False)
        Chem.perceiveHybridizationStates(mol, False)
        Chem.perceiveSSSR(mol, False)
        Chem.setRingFlags(mol, False)
        Chem.setAromaticityFlags(mol, False)
        Chem.perceiveSybylAtomTypes(mol, False)
        Chem.calcTopologicalDistanceMatrix(mol, False)
        Chem.perceivePiElectronSystems(mol, False)

        MolProp.calcPEOEProperties(mol, False)
        MolProp.calcMHMOProperties(mol, False)

        ForceField.perceiveMMFF94AromaticRings(mol, False)
        ForceField.assignMMFF94AtomTypes(mol, False, False)
        ForceField.assignMMFF94BondTypeIndices(mol, False, False)
        ForceField.calcMMFF94AtomCharges(mol, False, False)

    @staticmethod
    def extract_structure_data(mol: Chem.Molecule) -> dict:
        """Extract structure data from a molecule."""
        struct_data = Chem.getStructureData(mol) if Chem.hasStructureData(mol) else []
        data_dict = {}

        for entry in struct_data:
            if "soms" in entry.header:
                data_dict["soms"] = ast.literal_eval(entry.data)

        return data_dict


class DescriptorGenerator:
    """Class for generating descriptors for a given atom in a molecule."""

    def __init__(self, radius):
        """Initialize the DescriptorGenerator class."""
        self.radius = radius

    def generate_descriptors(
        self, ctr_atom: Chem.Atom, molgraph: Chem.MolecularGraph
    ) -> tuple[list[str], np.ndarray]:
        """Generate descriptors for a given atom in a molecule.

        Args:
            ctr_atom (Chem.Atom): center atom
            molgraph (Chem.MolecularGraph): molecular graph

        Returns:
            tuple: (descriptor_names, descriptors)
        """
        properties_dict = {
            Chem.getSybylType: "AtomType",
            MolProp.getHeavyAtomCount: "AtomDegree",
            MolProp.getHybridPolarizability: "HybridPolarizability",
            MolProp.getVSEPRCoordinationGeometry: "VSEPRgeometry",
            MolProp.calcExplicitValence: "AtomValence",
            MolProp.calcEffectivePolarizability: "EffectivePolarizability",
            MolProp.getPEOESigmaCharge: "SigmaCharge",
            ForceField.getMMFF94Charge: "MMFF94Charge",
            MolProp.calcPiElectronegativity: "PiElectronegativity",
            MolProp.getPEOESigmaElectronegativity: "SigmaElectronegativity",
            MolProp.calcInductiveEffect: "InductiveEffect",
        }

        fs, names = zip(*properties_dict.items())
        descriptor_names = [
            f"{prefix}_{Chem.getSybylAtomTypeString(i)}_{j}"
            for prefix in names
            for j in range(self.radius + 1)
            for i in SYBYL_ATOM_TYPE_IDX_CDPKIT
        ]
        descr = np.zeros(
            (len(properties_dict), len(SYBYL_ATOM_TYPE_IDX_CDPKIT) * (self.radius + 1)),
            dtype=float,
        )

        env = Chem.Fragment()
        Chem.getEnvironment(ctr_atom, molgraph, self.radius, env)

        for atom in env.atoms:
            sybyl_type = Chem.getSybylType(atom)
            if sybyl_type not in SYBYL_ATOM_TYPE_IDX_CDPKIT:
                continue

            position = SYBYL_ATOM_TYPE_IDX_CDPKIT.index(sybyl_type)
            top_dist = Chem.getTopologicalDistance(ctr_atom, atom, molgraph)
            descr[0, (top_dist * len(SYBYL_ATOM_TYPE_IDX_CDPKIT) + position)] += 1

            for i, f_, name in zip(range(len(fs)), fs, names):
                if name == "AtomType":
                    continue
                prop = (
                    f_(atom)
                    if name in ["SigmaCharge", "MMFF94Charge", "SigmaElectronegativity"]
                    else f_(atom, molgraph)
                )
                descr[
                    i, (top_dist * len(SYBYL_ATOM_TYPE_IDX_CDPKIT) + position)
                ] += prop

        for i in range(1, len(fs)):
            # averaging property and when divide by 0 give 0
            descr[i, :] = np.divide(
                descr[i, :],
                descr[0, :],
                out=np.zeros_like(descr[i, :]),
                where=descr[0, :] != 0,
            )

        max_top_dist, max_distance_center = self._calculate_distances(
            ctr_atom,
            molgraph,
        )
        descriptor_names += [
            "longestMaxTopDistinMolecule",
            "highestMaxTopDistinMatrixRow",
            "diffSPAN",
            "refSPAN",
        ]
        descr = descr.flatten()

        if max_top_dist == 0:
            descr = np.append(
                descr,
                [
                    max_top_dist,
                    max_distance_center,
                    max_top_dist - max_distance_center,
                    0,
                ],
            )
        else:
            descr = np.append(
                descr,
                [
                    max_top_dist,
                    max_distance_center,
                    max_top_dist - max_distance_center,
                    max_distance_center / max_top_dist,
                ],
            )

        descr = descr.round(4)
        return descriptor_names, descr

    def _calculate_distances(
        self,
        ctr_atom: Chem.Atom,
        molgraph: Chem.MolecularGraph,
    ) -> tuple:
        """Calculate the maximum topological distance across \
            the whole molecule and the maximum distance from the center atom.

        Args:
            ctr_atom (Chem.Atom): center atom
            molgraph (Chem.MolecularGraph): molecular graph

        Returns:
            tuple: (max_top_dist, max_distance_center)
        """
        max_top_dist = max(
            Chem.getTopologicalDistance(atom1, atom2, molgraph)
            for atom1 in molgraph.atoms
            for atom2 in molgraph.atoms
        )
        max_distance_center = max(
            Chem.getTopologicalDistance(ctr_atom, atom, molgraph)
            for atom in molgraph.atoms
        )
        return max_top_dist, max_distance_center


class FAMEDescriptors:
    """Class for computing FAME descriptors for a given molecule."""

    def __init__(self, radius):
        """Initialize the FAMEDescriptors class."""
        self.radius = radius
        self.descriptor_generator = DescriptorGenerator(radius)

    def _process_molecule(
        self, mol: Chem.Molecule, has_soms: bool
    ) -> tuple[tuple, dict]:
        """Process a molecule and generate descriptors for each atom.

        Args:
            mol (Chem.Molecule): molecule
            has_soms (bool): whether the input file contains site of metabolism labels

        Returns:
            tuple: (descriptor_names, property_dict)
        """
        data_dict = MoleculeProcessor.extract_structure_data(mol)
        mol_id = Chem.getName(mol)
        soms = data_dict.get("soms", [])

        MoleculeProcessor.perceive_mol(mol)
        property_dict = {}

        for atom in mol.atoms:
            if Chem.getSybylType(atom) == 24:
                continue  # Skip hydrogen atoms

            atom_id = mol.getAtomIndex(atom)
            (
                descriptor_names,
                descriptors,
            ) = self.descriptor_generator.generate_descriptors(atom, mol)
            som_label = 1 if has_soms and atom_id in soms else 0 if has_soms else None
            property_dict[(mol_id, atom_id)] = (som_label, descriptors)

        return descriptor_names, property_dict

    def compute_fame_descriptors(
        self, in_file: str, out_folder: str, has_soms: bool
    ) -> tuple[np.ndarray, list[str], np.ndarray, np.ndarray | None, np.ndarray]:
        """Compute FAME descriptors for a given molecule.

        Args:
            in_file (str): input file path
            out_folder (str): output folder path
            has_soms (bool): whether the input file contains site of metabolism labels

        Returns:
            tuple: (mol_num_ids, mol_ids, atom_ids, som_labels, descriptors_lst)
        """
        in_file_path, out_folder_path = Path(in_file), Path(out_folder)
        out_not_calculated, out_descriptors = self._get_output_file_paths(
            in_file_path, out_folder_path
        )

        if not os.path.exists(out_not_calculated) and not os.path.exists(
            out_descriptors
        ):
            return self._compute_and_save_descriptors(
                in_file, out_not_calculated, out_descriptors, has_soms
            )
        return self._read_precomputed_descriptors(out_descriptors, has_soms)

    def _get_output_file_paths(
        self, in_file_path: Path, out_folder_path: Path
    ) -> tuple[Path, Path]:
        """Generate output file paths for descriptors and not-calculated compounds."""
        out_not_calculated = (
            out_folder_path
            / f"{in_file_path.stem}_{self.radius}_not_calculated_cpds.csv"
        )
        out_descriptors = (
            out_folder_path / f"{in_file_path.stem}_{self.radius}_descriptors.csv"
        )
        return out_not_calculated, out_descriptors

    def _compute_and_save_descriptors(
        self,
        in_file: str,
        out_not_calculated: Path,
        out_descriptors: Path,
        has_soms: bool,
    ) -> tuple[np.ndarray, list[str], np.ndarray, np.ndarray | None, np.ndarray]:
        """Compute and save descriptors from molecules."""
        reader = Chem.FileSDFMoleculeReader(in_file)
        mol = Chem.BasicMolecule()

        mol_num_ids: list[int] = []
        mol_ids: list[str] = []
        atom_ids: list[int] = []
        som_labels: list[int] = []
        descriptors_lst: list[list[float]] = []

        with (
            open(out_not_calculated, "w", encoding="UTF-8") as f_not_calc_writer,
            open(out_descriptors, "w", encoding="UTF-8") as f_desc_writer,
        ):
            f_not_calc_csv = csv.writer(f_not_calc_writer)
            f_desc_csv = csv.writer(f_desc_writer)

            f_not_calc_csv.writerow(["sybyl_atom_type_id", "sybyl_atom_type", "mol_id"])

            i = 0
            try:
                while reader.read(mol):
                    try:
                        MoleculeProcessor.perceive_mol(mol)
                        if self._has_uncommon_element(mol, f_not_calc_csv):
                            continue

                        descriptor_names, property_dict = self._process_molecule(
                            mol, has_soms
                        )

                        if i == 0:
                            f_desc_csv.writerow(
                                ["mol_num_id", "mol_id", "atom_id", "som_label"]
                                + list(descriptor_names)
                            )
                        i += 1

                        for (mol_id, atom_id), (
                            som_label,
                            descriptors,
                        ) in property_dict.items():
                            mol_num_ids.append(i)
                            mol_ids.append(mol_id)
                            atom_ids.append(atom_id)
                            som_labels.append(som_label)
                            descriptors_lst.append(list(descriptors))

                            f_desc_csv.writerow(
                                [i, mol_id, atom_id, som_label] + list(descriptors)
                            )
                    except RuntimeError as e:
                        sys.exit(f"Error: processing of molecule failed:\n{e}")
            except RuntimeError as e:
                sys.exit(f"Error: reading molecule failed:\n{e}")

            return (
                np.array(mol_num_ids, int),
                mol_ids,
                np.array(atom_ids, int),
                np.array(som_labels, int) if has_soms else None,
                np.array(descriptors_lst, float),
            )

    def _has_uncommon_element(self, mol, f_not_calc) -> bool:
        """Check if a molecule contains uncommon elements and log them."""
        uncommon_element = False
        for atom in mol.atoms:
            atom_type = Chem.getSybylType(atom)
            if atom_type not in SYBYL_ATOM_TYPE_IDX_CDPKIT:
                f_not_calc.writerow([atom_type, Chem.getSybylAtomTypeString(atom_type)])
                uncommon_element = True
        return uncommon_element

    def _read_precomputed_descriptors(
        self, out_descriptors: Path, has_soms: bool
    ) -> tuple[np.ndarray, list[str], np.ndarray, np.ndarray | None, np.ndarray]:
        """Read previously computed descriptors from a file."""
        print(f"Reading pre-calculated descriptors from {out_descriptors}")

        mol_num_ids, mol_ids, atom_ids, som_labels, descriptors_lst = [], [], [], [], []

        with open(out_descriptors, "r", encoding="UTF-8") as f:
            next(f)  # Skip header
            for line in f:
                parts = line.strip().split(",")
                mol_num_id, mol_id, atom_id, som_label = parts[:4]
                descriptors = list(map(float, parts[4:]))
                mol_num_ids.append(mol_num_id)
                mol_ids.append(mol_id)
                atom_ids.append(atom_id)
                som_labels.append(som_label)
                descriptors_lst.append(descriptors)

        return (
            np.array(mol_num_ids, int),
            mol_ids,
            np.array(atom_ids, int),
            np.array(som_labels, int) if has_soms else None,
            np.array(descriptors_lst, float),
        )



"""Applies a trained re-implementation of the FAME3 model to unlabeled data.

This script saves the per-atom predictions to a CSV file.
The radius of the atom environment is not part of the hyperparameter search, \
    but can be set by changing the radius argument. Default is 5.
The decision threshold can be changed by changing the threshold argument. Default is 0.3.
The script also computes FAME scores if the -fs flag is set.
"""

import argparse
import csv
import glob
import os
import sys
from datetime import datetime

import pandas as pd
from joblib import load

from fame3r import FAMEDescriptors, FAMEScores


def parse_arguments() -> argparse.Namespace:
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description="Applies a trained re-implementation of the FAME3 model to unlabeled data"
    )

    parser.add_argument(
        "-i",
        dest="input_file",
        required=True,
        metavar="<Input data file>",
        help="Input data file",
    )
    parser.add_argument(
        "-m",
        dest="model_folder",
        required=True,
        metavar="<Model folder>",
        help="Model folder",
    )
    parser.add_argument(
        "-o",
        dest="out_folder",
        required=True,
        metavar="<Output folder>",
        help="Output location",
    )
    parser.add_argument(
        "-r",
        dest="radius",
        required=False,
        metavar="<radius>",
        default=5,
        help="Max. atom environment radius in number of bonds",
        type=int,
    )
    parser.add_argument(
        "-t",
        dest="threshold",
        required=False,
        metavar="<binary decision threshold>",
        default=0.3,
        help="Binary decision threshold",
        type=float,
    )
    parser.add_argument(
        "-fs",
        dest="compute_fame_scores",
        action="store_true",
        help="Compute FAME scores (optional)",
    )

    return parser.parse_args()


def main():
    """Application entry point."""
    start_time = datetime.now()

    args = parse_arguments()
    print(f"Radius: {args.radius}")

    if not os.path.exists(args.out_folder):
        os.makedirs(args.out_folder)
        print("The new output folder is created.")

    print("Computing descriptors...")
    descriptors_generator = FAMEDescriptors(args.radius)
    mol_num_ids, mol_ids, atom_ids, _, descriptors = descriptors_generator.compute_fame_descriptors(
        args.input_file, args.out_folder, has_soms=False
    )

    print(f"Data: {len(set(mol_num_ids))} molecules")

    print("Loading model...")
    clf = load(os.path.join(args.model_folder, "model.joblib"))

    print("Predicting SOMs...")
    predictions = clf.predict_proba(descriptors)[:, 1]
    predictions_binary = (predictions > args.threshold).astype(int)

    fame_scores = None
    if args.compute_fame_scores:
        print("Computing FAME scores...")
        csv_files = glob.glob(os.path.join(args.model_folder, "*descriptors.csv"))
        if len(csv_files) == 1:
            train_descriptors_df = pd.read_csv(csv_files[0])
        else:
            raise FileNotFoundError(
                f"Expected one CSV file ending with 'descriptors.csv', but found {len(csv_files)}."
            )
        train_descriptors = train_descriptors_df.drop(
            columns=["mol_num_id", "mol_id", "atom_id", "som_label"]
        ).values
        fame_scores_generator = FAMEScores(train_descriptors)
        fame_scores = fame_scores_generator.compute_fame_scores(descriptors)

    predictions_file = os.path.join(args.out_folder, "predictions.csv")
    with open(predictions_file, "w", encoding="UTF-8", newline="") as file:
        writer = csv.writer(file)
        headers = ["mol_id", "atom_id", "y_prob", "y_pred"]
        if args.compute_fame_scores:
            headers.append("fame_score")
        writer.writerow(headers)

        for i in range(len(mol_ids)):
            row = [mol_ids[i], atom_ids[i], predictions[i], predictions_binary[i]]
            if args.compute_fame_scores:
                row.append(fame_scores[i])
            writer.writerow(row)

    print(f"Predictions saved to {predictions_file}")
    print("Finished in:", datetime.now() - start_time)

    sys.exit(0)


if __name__ == "__main__":
    main()



"""Trains a re-implementation of the FAME3 model.

This script trains a random forest classifier using the FAME descriptors.
The model is saved in the output folder as a joblib file.
The hyperparameters of the model and the radius of the atom environment \
    are not optimized in this script (see cv_hp_search.py). \
        The hyperparameters can be set in the RandomForestClassifier constructor. \
            The radius can be set by changing the radius command line argument (default is 5).

"""

import argparse
import os
import sys
from datetime import datetime

from joblib import dump
from sklearn.ensemble import RandomForestClassifier

from fame3r import FAMEDescriptors


def parse_arguments() -> argparse.Namespace:
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description="Trains a re-implementation of the FAME3 model."
    )

    parser.add_argument(
        "-i",
        dest="input_file",
        required=True,
        metavar="<training data file>",
        help="Training data file",
    )
    parser.add_argument(
        "-o",
        dest="out_folder",
        required=True,
        metavar="<Output folder>",
        help="Output location",
    )
    parser.add_argument(
        "-r",
        dest="radius",
        required=False,
        metavar="<radius>",
        default=5,
        help="Max. atom environment radius in number of bonds",
        type=int,
    )

    return parser.parse_args()


def main():
    """Application entry point."""
    start_time = datetime.now()

    args = parse_arguments()
    print(f"Radius: {args.radius}")

    if not os.path.exists(args.out_folder):
        os.makedirs(args.out_folder)
        print("The new output folder is created.")

    print("Computing descriptors...")

    descriptors_generator = FAMEDescriptors(args.radius)
    (
        mol_num_ids,
        _mol_ids,
        _atom_ids,
        som_labels,
        descriptors,
    ) = descriptors_generator.compute_fame_descriptors(
        args.input_file, args.out_folder, has_soms=True
    )

    print(f"Training data: {len(set(mol_num_ids))} molecules")

    print("Training model...")
    clf = RandomForestClassifier(
        n_estimators=100,
        max_features="sqrt",
        class_weight="balanced",
        random_state=42,
    )
    clf.fit(descriptors, som_labels)

    print("Saving model...")
    dump(clf, os.path.join(args.out_folder, "model.joblib"))

    print("Finished in:", datetime.now() - start_time)

    sys.exit(0)


if __name__ == "__main__":
    main()
